///////////////////////////////////////////////////////////////////////////////
// File: HTTPSocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2016 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "HTTPSocketServer.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/IIndexedOpaqueUserData.h"

#include "JetByteTools/IOTools/Buffer.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/IAllocateSequencedStreamSockets.h"


#pragma hdrstop

#include "JetByteTools/HTTPTools/IManageResources.h"
#include "JetByteTools/HTTPTools/HTTPConnection.h"

#define DEBUG_HTTP_SERVER

#ifdef DEBUG_HTTP_SERVER
#define DEBUGGING_CONNECTIONS(_e) JetByteTools::Core::OutputEx(_e)
#else
#define DEBUGGING_CONNECTIONS(_e)
#endif

#ifdef DEBUG_HTTP_SERVER_FLOW_CONTROL
#define DEBUGGING_FLOW_CONTROL(e) JetByteTools::Core::OutputEx(e);
#else
#define DEBUGGING_FLOW_CONTROL(e)
#endif

#if defined(DEBUG_HTTP_REQUEST_PERFORMANCE) || defined(DEBUG_HTTP_SERVER)
static const JetByteTools::Core::_tstring s_sep(_T(" - "));
#define DEBUGGING_REQUEST_PERF(e, s) JetByteTools::Core::OutputEx(e + s_sep + JetByteTools::Core::PointerToString(&s));
#else
#define DEBUGGING_REQUEST_PERF(e, s)
#endif

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::Output;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::CException;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::ToString;

using JetByteTools::IO::IAllocateBuffers;
using JetByteTools::IO::IBuffer;
using JetByteTools::IO::IIOPool;
using JetByteTools::IO::IBufferChain;
using JetByteTools::IO::CSmartBuffer;

using JetByteTools::Socket::CFilteringStreamSocketServerEx;
using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateSequencedStreamSockets;
using JetByteTools::Socket::IFullAddress;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::NoZeroByteRead;
using JetByteTools::Socket::CFlowControlStreamSocketConnectionFilter;
using JetByteTools::Socket::SocketBufferSize;
using JetByteTools::Socket::IMonitorStreamSocketFlowControl;

using JetByteTools::HTTP::IManageResources;
using JetByteTools::HTTP::CHTTPConnection;

using std::string;

///////////////////////////////////////////////////////////////////////////////
// CHTTPSocketServer
///////////////////////////////////////////////////////////////////////////////

CHTTPSocketServer::CHTTPSocketServer(
   IManageResources &resourceManager,
   const CFlowControlStreamSocketConnectionFilter::BufferCount maxBuffers,
   const CFlowControlStreamSocketConnectionFilter::PendingWriteCount maxPendingWrites,
   const CFlowControlStreamSocketConnectionFilter::Percentage notificationPercentageStepSize,
   const SocketBufferSize recvBufferSize,
   const SocketBufferSize sendBufferSize,
   const IFullAddress &address,
   ListenBacklog listenBacklog,
   IIOPool &ioPool,
   IAllocateSequencedStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator)
   :  CFilteringStreamSocketServerEx(address, listenBacklog, *this, ioPool, socketAllocator, bufferAllocator, recvBufferSize, sendBufferSize, NoZeroByteRead),
      m_resourceManager(resourceManager),
      m_flowControlFilter(*this, *this, socketAllocator, bufferAllocator, maxBuffers, *this, maxPendingWrites, notificationPercentageStepSize),
      m_connectionIndex(socketAllocator.RequestUserDataSlot(_T("PCMStreamer::m_connectionIndex")))
{

}

CHTTPSocketServer::~CHTTPSocketServer()
{
   try
   {
      WaitForShutdownToComplete();
   }
   JETBYTE_CATCH_AND_LOG_ALL_IN_DESTRUCTORS_IF_ENABLED
}

void CHTTPSocketServer::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   DEBUGGING_REQUEST_PERF(_T("CHTTPSocketServer::OnConnectionEstablished()"), socket);

   CHTTPConnection *pConnection = new CHTTPConnection(socket, m_resourceManager);

   socket.SetUserPointer(m_connectionIndex, pConnection);

   socket.Read();
}

void CHTTPSocketServer::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   DEBUGGING_REQUEST_PERF(_T("CHTTPSocketServer::OnSocketReleased()"), socket);

   CHTTPConnection *pConnection = reinterpret_cast<CHTTPConnection *>(userData.GetUserPointer(m_connectionIndex));

   delete pConnection;
}

void CHTTPSocketServer::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   try
   {
      DEBUGGING_REQUEST_PERF(_T("CHTTPSocketServer::ReadCompleted()"), socket);

      const IBuffer::BufferSize dataSize = buffer->GetUsed();

      const BYTE *pData = buffer->GetMemory();

      const string data(reinterpret_cast<const char *>(pData), dataSize);

      DEBUGGING_CONNECTIONS(_T("Incoming HTTP Data:\r\n") + CStringConverter::AtoT(data));

      //DEBUGGING_CONNECTIONS(_T("Incoming HTTP Data:\r\n") + DumpData(buffer->GetMemory(), buffer->GetUsed(), 60, true));

      CHTTPConnection *pConnection = reinterpret_cast<CHTTPConnection *>(socket.GetUserPointer(m_connectionIndex));

      pConnection->OnDataReceived(buffer);
   }
   catch(const CException &e)
   {
      Output(_T("CHTTPSocketServer::ReadCompleted - Exception - ") + e.GetDetails());
      socket.AbortConnection();
   }
   catch(...)
   {
      Output(_T("CHTTPSocketServer::ReadCompleted - Unexpected exception"));
      socket.AbortConnection();
   }
}

void CHTTPSocketServer::OnConnectionClientClose(
   IStreamSocket &socket)
{
   DEBUGGING_CONNECTIONS(_T("OnConnectionClientClose"));

   CHTTPConnection *pConnection = reinterpret_cast<CHTTPConnection *>(socket.GetUserPointer(m_connectionIndex));

   (void)pConnection;

   socket.Shutdown();

//   pConnection->OnConnectionClientClose();
}

bool CHTTPSocketServer::OnListFull(
   IStreamSocket &socket,
   const ConnectionId id,
   IMonitorStreamSocketFlowControl &monitor,
   const IBuffer &newBuffer,
   IBufferChain &buffers,
   IBufferChain &discardedBuffers)
{
   (void)id;
   (void)monitor;
   (void)newBuffer;
   (void)buffers;
   (void)discardedBuffers;

   DEBUGGING_FLOW_CONTROL(_T("CHTTPSocketServer::OnListFull"));

   CHTTPConnection *pConnection = reinterpret_cast<CHTTPConnection *>(socket.GetUserPointer(m_connectionIndex));

   pConnection->PauseOutput();

   // return true to add the buffer to the queue anyway and false to discard.
   // DO NOT ADD THE NEW BUFFER INTO THE DISCARDED BUFFERS LIST IF YOU DISCARD!!!

   return true;
}

void CHTTPSocketServer::OnListStatus(
   IStreamSocket &socket,
   const DWORD percentageUsed)
{
   (void)socket;
   (void)percentageUsed;

   DEBUGGING_FLOW_CONTROL(_T("CHTTPSocketServer::OnListStatus - ") + ToString(percentageUsed) + _T("%"));
}

void CHTTPSocketServer::OnListEmpty(
   IStreamSocket &socket)
{
   DEBUGGING_FLOW_CONTROL(_T("CHTTPSocketServer::OnListEmpty"));

   CHTTPConnection *pConnection = reinterpret_cast<CHTTPConnection *>(socket.GetUserPointer(m_connectionIndex));

   pConnection->ResumeOutput();
}

// Implement IMonitorStreamSocketFlowControl

IMonitorStreamSocketFlowControl::ConnectionId CHTTPSocketServer::OnConnectionEstablished(
   IStreamSocket &socket)
{
   IMonitorStreamSocketFlowControl::ConnectionId id = reinterpret_cast<IMonitorStreamSocketFlowControl::ConnectionId>(&socket);

   //DEBUGGING_FLOW_CONTROL(_T("OnConnectionEstablished: ") + ToString(id));

   return id;
}

void CHTTPSocketServer::OnWritePending(
   const ConnectionId id)
{
   (void)id;

   //DEBUGGING_FLOW_CONTROL(_T("OnWritePending: ") + ToString(id));
}

void CHTTPSocketServer::OnWriteComplete(
   const ConnectionId id)
{
   (void)id;

   //DEBUGGING_FLOW_CONTROL(_T("OnWriteComplete: ") + ToString(id));
}

bool CHTTPSocketServer::OnWriteCompletionError(
   const ConnectionId id,
   IStreamSocket &socket,
   IBuffer &buffer,
   const DWORD lastError)
{
   (void)id;
   (void)socket;
   (void)buffer;
   (void)lastError;

   //DEBUGGING_FLOW_CONTROL(_T("OnWriteCompletionError: ") + ToString(id) + _T(" - ") + ToString(lastError));

   // Return true if we can continue to send data and false if we should purge the flow control queue.

   return false;
}

void CHTTPSocketServer::OnBufferDelayed(
   const ConnectionId id,
   IBuffer &buffer)
{
   (void)buffer;
   (void)id;

   //DEBUGGING_FLOW_CONTROL(_T("OnBufferDelayed: ") + ToString(id));
}

void CHTTPSocketServer::OnBufferSent(
   const ConnectionId id,
   IBuffer &buffer)
{
   (void)id;
   (void)buffer;

   //DEBUGGING_FLOW_CONTROL(_T("OnBufferSent: ") + ToString(id));
}

void CHTTPSocketServer::OnBufferDiscarded(
   const ConnectionId id,
   IBuffer &buffer,
   const bool wasDelayed)
{
   (void)id;
   (void)buffer;
   (void)wasDelayed;

   //DEBUGGING_FLOW_CONTROL(_T("OnBufferDiscarded: ") + ToString(id) + (wasDelayed ? _T(" was queued") : _T("")));
}

void CHTTPSocketServer::OnConnectionComplete(
   const ConnectionId id)
{
   (void)id;

   //DEBUGGING_FLOW_CONTROL(_T("OnConnectionComplete: ") + ToString(id));
}

void CHTTPSocketServer::OnConnectionTerminated(
   const ConnectionId id)
{
   (void)id;

   //DEBUGGING_FLOW_CONTROL(_T("OnConnectionTerminated: ") + ToString(id));
}

///////////////////////////////////////////////////////////////////////////////
// End of file: HTTPSocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
