///////////////////////////////////////////////////////////////////////////////
// File: StreamSocketConnectionManager.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2005 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "StreamSocketConnectionManager.h"
#include "TestMessages.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/WaitableCounter.h"
#include "JetByteTools/CoreTools/StringConverter.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/IAllocateSequencedStreamSockets.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::Output;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::CException;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::ToStringA;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::IQueueTimers;
using JetByteTools::Core::ErrorCodeToErrorMessage;
using JetByteTools::Core::ErrorCodeToErrorMessageA;
using JetByteTools::Core::ILockableObject;

using JetByteTools::IO::IIOPool;
using JetByteTools::IO::IBuffer;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IAllocateBuffers;

using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateSequencedStreamSockets;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ShutdownSend;
using JetByteTools::Socket::ShutdownReceive;
using JetByteTools::Socket::NoZeroByteRead;
using JetByteTools::Socket::ConnectionClosureReason;
using JetByteTools::Socket::ConnectionAbortedMask;
using JetByteTools::Socket::GetClosureReasonAsString;

using std::cout;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// CStreamSocketConnectionManager
///////////////////////////////////////////////////////////////////////////////

CStreamSocketConnectionManager::CStreamSocketConnectionManager(
   const Milliseconds readTimeout,
   IIOPool &pool,
   IAllocateSequencedStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator)
   :  JetByteTools::Socket::CFilteringStreamSocketConnectionManager(*this, pool, socketAllocator, bufferAllocator, NoZeroByteRead),
      m_timerHandleIndex(socketAllocator.RequestUserDataSlot(_T("CStreamSocketConnectionManager - m_timerHandleIndex"))),
      m_userDataIndex(socketAllocator.RequestUserDataSlot(_T("CStreamSocketConnectionManager - m_userDataIndex"))),
      m_timeoutFilter(*this, m_timerQueue),
      m_readTimeout(readTimeout)
{

}

CStreamSocketConnectionManager::~CStreamSocketConnectionManager()
{
   WaitForShutdownToComplete();
}

void CStreamSocketConnectionManager::OnPreOutgoingConnect(
   IIndexedOpaqueUserData &userData,
   const IAddress & /*address*/,
   const void *pUserData)
{
   const auto pTestMessages = const_cast<CTestMessages *>(static_cast<const CTestMessages*>(pUserData));

   userData.SetUserPointer(m_userDataIndex, pTestMessages);
}

void CStreamSocketConnectionManager::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   pTestMessages->ConnectionComplete(true);

   const auto timerHandle = m_timerQueue.CreateTimer();

   socket.SetUserData(m_timerHandleIndex, timerHandle);

   if (m_readTimeout != 0)
   {
      m_timeoutFilter.SetReadTimeout(socket, m_readTimeout, *this, 1);
   }

   socket.TryRead();

   SendData(socket);
}

void CStreamSocketConnectionManager::OnOutgoingConnectionFailed(
   IStreamSocket &socket,
   const IAddress & /*address*/,
   const DWORD lastError)
{
   {
      ILockableObject::Owner lock(m_lock);

      m_connectionErrors[lastError]++;
   }

   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   socket.SetUserData(m_timerHandleIndex, IQueueTimers::InvalidHandleValue);

   pTestMessages->ConnectionComplete(false);

   socket.SetUserPointer(m_userDataIndex, nullptr);
}

void CStreamSocketConnectionManager::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   const auto timerHandle = userData.GetUserData(m_timerHandleIndex);

   if (timerHandle != IQueueTimers::InvalidHandleValue)
   {
      if (m_timerQueue.DestroyTimer(timerHandle))
      {
         OnError(_T("CStreamSocketConnectionManager::OnSocketReleased() - timer was set when destroyed, socket ref count bug!"));
      }
   }
}

void CStreamSocketConnectionManager::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   try
   {
      ProcessData(socket, buffer);

      socket.TryRead(buffer);
   }
   catch(const CException &e)
   {
      OutputEx(_T("ReadCompleted - Exception - ") + e.GetDetails());
      socket.AbortConnection();
   }
   catch(...)
   {
      OutputEx(_T("ReadCompleted - Unexpected exception"));
      socket.AbortConnection();
   }
}

void CStreamSocketConnectionManager::OnConnectionReset(
   IStreamSocket & /*socket*/,
   const DWORD lastError)
{
   ILockableObject::Owner lock(m_lock);

   m_connectionResetErrors[lastError]++;
}

void CStreamSocketConnectionManager::OnError(
   const _tstring &message)
{
   ILockableObject::Owner lock(m_lock);

   m_errors[message]++;
}

void CStreamSocketConnectionManager::OnConnectionClosure(
   IStreamSocket &socket,
   const ConnectionClosureReason reason)
{
   const auto timerHandle = socket.GetUserData(m_timerHandleIndex);

   socket.SetUserData(m_timerHandleIndex, IQueueTimers::InvalidHandleValue);

   if (timerHandle != IQueueTimers::InvalidHandleValue)
   {
      m_timerQueue.DestroyTimerWithRefCountedUserData(timerHandle, &socket);
   }

   if (reason & ConnectionAbortedMask)
   {
      OnError(GetClosureReasonAsString(reason));
   }

   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   if (pTestMessages)
   {
      pTestMessages->MessagesComplete(!(reason & ConnectionAbortedMask));
   }
}

void CStreamSocketConnectionManager::ProcessData(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   if (pTestMessages)
   {
      try
      {
         auto messageComplete = false;

         if (pTestMessages->ValidateReceivedData(buffer->GetMemory(), buffer->GetUsed(), messageComplete))
         {
            if (messageComplete && pTestMessages->SendAfterRecv())
            {
               SetSendTimer(socket, pTestMessages->GetTransferRateMillis());
            }
         }
         else if (messageComplete)
         {
            const auto holdConnections = pTestMessages->HoldConnections();

            if (!holdConnections)
            {
               socket.Shutdown(ShutdownSend);
            }
            else
            {
               pTestMessages->MessagesComplete(true);
            }
         }
      }
      catch (CException &e)
      {
         ILockableObject::Owner lock(m_lock);

         m_errors[e.GetWhat()]++;

         pTestMessages->MessagesComplete(false);

         throw;
      }
   }

   buffer->Empty();
}

void CStreamSocketConnectionManager::OnTimer(
   IQueueTimers::Timer::UserData userData)
{
   auto &socket = *reinterpret_cast<IStreamSocket *>(userData);

   SendData(socket);

   socket.Release();
}

void CStreamSocketConnectionManager::OnTimer(
   IStreamSocket &socket,
   TimerCallback::UserData /*userData*/)
{
   OnError(_T("Read timeout"));

   socket.AbortConnection();
}

void CStreamSocketConnectionManager::SendData(
   IStreamSocket &socket)
{
   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   if (pTestMessages)
   {
      if (pTestMessages->WriteSome(socket))
      {
         if (!pTestMessages->SendAfterRecv())
         {
            SetSendTimer(socket, pTestMessages->GetTransferRateMillis());
         }
      }
      else if (!pTestMessages->HoldConnections())
      {
         socket.Shutdown(ShutdownSend);
      }
   }
}

void CStreamSocketConnectionManager::SetSendTimer(
   IStreamSocket &socket,
   const Milliseconds timeout)
{
   const auto timerHandle = socket.GetUserData(m_timerHandleIndex);

   m_timerQueue.SetTimerWithRefCountedUserData(timerHandle, *this, timeout, &socket);
}

bool CStreamSocketConnectionManager::DumpErrors() const
{
   auto ok = true;

   if (!DumpErrors(_T("Connection establishment errors"), m_connectionErrors))
   {
      ok = false;
   }

   if (!DumpErrors(_T("Connections reset"), m_connectionResetErrors))
   {
      ok = false;
   }

   if (!m_errors.empty())
   {
      ok = false;

      cout << "Other errors" << endl;

      for (const auto &error : m_errors)
      {
         cout << ToStringA(error.second) << " - " << CStringConverter::TtoA(error.first) << endl;
      }
   }

   return ok;
}

bool CStreamSocketConnectionManager::DumpErrors(
   const _tstring &errorType,
   const LastErrors &lastErrors) const
{
   auto ok = true;

   if (!lastErrors.empty())
   {
      ok = false;

      cout << CStringConverter::TtoA(errorType) << endl;

      for (const auto &lastError : lastErrors)
      {
         cout << ToStringA(lastError.second) << " - " << ErrorCodeToErrorMessageA(lastError.first) << endl;
      }
   }

   return ok;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: StreamSocketConnectionManager.cpp
///////////////////////////////////////////////////////////////////////////////
